# -*- coding: utf-8 -*-
"default plugin for value: set it in a simple dictionary"
# Copyright (C) 2013-2014 Team tiramisu (see AUTHORS for all contributors)
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ____________________________________________________________
from ..util import Cache
from tiramisu.setting import undefined


class Values(Cache):
    __slots__ = ('_values', '_informations', '__weakref__')

    def __init__(self, storage):
        """init plugin means create values storage
        """
        self._values = (tuple(), tuple(), tuple(), tuple())
        self._informations = {}
        # should init cache too
        super(Values, self).__init__(storage)

    # value
    def setvalue(self, path, value, owner, index):
        """set value for a path
        a specified value must be associated to an owner
        """
        values = []
        vidx = None

        def _setvalue_info(nb, idx, value, vidx):
            lst = list(self._values[nb])
            if idx is None:
                if index is None or nb == 0:
                    lst.append(value)
                else:
                    lst.append([value])
            else:
                if index is None or nb == 0:
                    lst[idx] = value
                else:
                    if nb == 1:
                        try:
                            vidx = lst[idx].index(index)
                        except ValueError:
                            vidx = None
                    if vidx is None:
                        lst[idx].append(value)
                    elif nb != 1:
                        lst[idx][vidx] = value
            values.append(tuple(lst))
            return vidx
        try:
            idx = self._values[0].index(path)
        except ValueError:
            idx = None
        vidx = _setvalue_info(0, idx, path, vidx)
        vidx = _setvalue_info(1, idx, index, vidx)
        vidx = _setvalue_info(2, idx, value, vidx)
        _setvalue_info(3, idx, owner, vidx)
        self._values = tuple(values)

    def getvalue(self, path, index=None):
        """get value for a path
        return: only value, not the owner
        """
        return self._getvalue(path, 2, index)

    def hasvalue(self, path, index=None):
        """if path has a value
        return: boolean
        """
        return path in self._values[0]

    def resetvalue(self, path):
        """remove value means delete value in storage
        """
        def _resetvalue(nb):
            lst = list(self._values[nb])
            lst.pop(idx)
            values.append(tuple(lst))
        values = []
        idx = self._values[0].index(path)
        _resetvalue(0)
        _resetvalue(1)
        _resetvalue(2)
        _resetvalue(3)
        self._values = tuple(values)

    def get_modified_values(self):
        """return all values in a dictionary
        example: {'path1': (owner, 'value1'), 'path2': (owner, 'value2')}
        """
        values = {}
        for idx, path in enumerate(self._values[0]):
            values[path] = (self._values[3][idx], self._values[2][idx])
        return values

    # owner
    def setowner(self, path, owner, index=None):
        """change owner for a path
        """
        idx = self._values[0].index(path)
        if isinstance(self._values[3][idx], list):
            if index is None:
                raise ValueError('list but no index')
            owner = list(self._values[3][idx])[index] = owner
        elif index is not None:
            raise ValueError('index set but not a list')
        lst = list(self._values[3])
        lst[idx] = owner
        values = list(self._values)
        values[3] = tuple(lst)
        self._values = tuple(values)

    def get_max_length(self, path):
        try:
            idx = self._values[0].index(path)
        except ValueError:
            return 0
        return max(self._values[1][idx]) + 1

    def getowner(self, path, default, index=None, only_default=False):
        """get owner for a path
        return: owner object
        """
        if index is None:
            if only_default:
                if path in self._values[0]:
                    return None
                else:
                    return default
            val = self._getvalue(path, 3, index)
            if val is None:
                return default
            return val
        else:
            value = self._getvalue(path, 3, index)
            if value is None:
                return default
            else:
                return value

    def _getvalue(self, path, nb, index):
        """
        _values == ((path1, path2), ((value1_1, value1_2), value2), ((owner1_1, owner1_2), owner2), ((idx1_1, idx1_2), None))
        """
        try:
            idx = self._values[0].index(path)
        except ValueError:
            value = None
        else:
            if isinstance(self._values[1][idx], list):
                if index is None:
                    raise ValueError('list but no index')
            elif index is not None:
                raise ValueError('index set but not a list')

            if self._values[1][idx] is None:
                if index is None:
                    value = self._values[nb][idx]
                else:
                    value = self._values[nb][idx][index]
            else:
                if index is not None:
                    try:
                        subidx = self._values[1][idx].index(index)
                        value = self._values[nb][idx][subidx]
                    except ValueError:
                        value = None
                else:
                    value = []
                    for i in xrange(0, max(self._values[1][idx])):
                        if i in self._values[1][idx]:
                            value.append(self._values[nb][idx][self._values[1][idx].index(i)])
                        else:
                            value.append(undefined)
        return value

    def set_information(self, key, value):
        """updates the information's attribute
        (which is a dictionary)

        :param key: information's key (ex: "help", "doc"
        :param value: information's value (ex: "the help string")
        """
        self._informations[key] = value

    def get_information(self, key):
        """retrieves one information's item

        :param key: the item string (ex: "help")
        """
        if key in self._informations:
            return self._informations[key]
        else:  # pragma: optional cover
            raise ValueError("not found")
